IMPLEMENTATION MODULE GeoRgn;
  (*
    Copyright (C) 1993-1995, Alex Iakovlev & Dimon Maslov
    **********************************************************************
    A.Iakovlev  Ver  1.0  Initial version                     February 90
    A.Iakovlev  Ver  2.0  Type Status, procedure Exist        February 91
    A.Iakovlev  Ver  3.0  Procedure XRectUpdate               February 91
    A.Iakovlev  Ver  4.0  Use module Res                      August 92
    A.Iakovlev  Ver  5.0  Renaming                            June 93
    A.Iakovlev  Ver  5.1  Use module Mem for heap control     October 94
    A.Iakovlev  Ver  6.0  No use module Res                   March 95
    **********************************************************************
  *)

  IMPORT Geo, DrvMem, SrvObj;

  CONST MaxRectsNmb = 100;

  TYPE
    REGION = POINTER TO RegionDescriptor;
    RectsArray = ARRAY [0 .. MaxRectsNmb - 1] OF Geo.Rectangle;
    RegionDescriptor =
      RECORD
        next: REGION;
        count: CARDINAL;
        rect: RectsArray;
      END;

  VAR topRgn: REGION;

  PROCEDURE New(r: Geo.Rectangle; VAR rgn: REGION): BOOLEAN;
    VAR
      newRgn: REGION;
    BEGIN
      IF DrvMem.Alloc(newRgn, SIZE(RegionDescriptor)) THEN
        WITH newRgn^ DO
          count := 1;
          rect[0] := r;
          (* -------------- include the region into the region chain *)
          next := topRgn;
        END;
        (* --------------------- normal end *)
        topRgn := newRgn;
        rgn := newRgn;
        RETURN TRUE;
      ELSE
        (* --------------- fail due to heap limitations *)
        RETURN FALSE;
      END;
    END New;

  PROCEDURE RectsNmb(rgn: REGION): CARDINAL;
    BEGIN
      RETURN rgn^.count;
    END RectsNmb;

  PROCEDURE XRect(rgn: REGION; r: Geo.Rectangle): BOOLEAN;
    VAR
      j: CARDINAL;
    BEGIN
      WITH rgn^ DO
        IF count > 0 THEN
          FOR j := 0 TO count - 1 DO
            IF Geo.RectXRect(rect[j], r) THEN
              RETURN TRUE;
            END;
          END;
        END;
      END;
      RETURN FALSE;
    END XRect;

  PROCEDURE XRectUpdate(rgn: REGION; r: Geo.Rectangle);
    VAR
      rect0: RectsArray;
      count0, j: CARDINAL;
    BEGIN
      WITH rgn^ DO
        (* ------------- copy initial region rectangles *)
        count0 := count;
        rect0 := rect;
        (* ------------------------------ compute the intersection *)
        count := 0;
        IF count0 > 0 THEN
          FOR j := 0 TO count0 - 1 DO
            IF Geo.RectXRectRes(rect0[j], r, rect[count]) THEN
              INC(count);
            END;
          END;
        END;
      END;
    END XRectUpdate;

  PROCEDURE XRectNew(rgn: REGION; r: Geo.Rectangle; VAR rgnRes: REGION): BOOLEAN;
    VAR
      newRgn: REGION;
      j: CARDINAL;
    BEGIN
      IF DrvMem.Alloc(newRgn, SIZE(RegionDescriptor)) THEN
        WITH newRgn^ DO
          (* -------------------- compute the intersection *)
          count := 0;
          IF rgn^.count > 0 THEN
            FOR j := 0 TO rgn^.count - 1 DO
              IF Geo.RectXRectRes(rgn^.rect[j], r, rect[count]) THEN
                INC(count);
              END;
            END;
          END;
          (* ------------------ include the region into the region chain *)
          next := topRgn;
        END;
        (* ------------- normal end *)
        topRgn := newRgn;
        rgnRes := newRgn;
        RETURN TRUE;
      ELSE
        (* ---------------- fail due to heap limitations *)
        RETURN FALSE;
      END;
    END XRectNew;

  PROCEDURE MinusRectUpdate(VAR rgn: REGION; r: Geo.Rectangle): BOOLEAN;
    VAR
      rect0: RectsArray;
      count0, j: CARDINAL;
    BEGIN
      WITH rgn^ DO
        (* ------------- store initial region rectangles *)
        count0 := count;
        rect0 := rect;
        (* ------------------------------ compute the difference *)
        count := 0;
        IF count0 > 0 THEN
          FOR j := 0 TO count0 - 1 DO
            count := Geo.RectMinusRect(rect0[j], r, rect, count, MaxRectsNmb - 1);
          END;
        END;
        (* -------------------- test the resulting region - TEMPORARILY !!! *)
        IF count > MaxRectsNmb THEN
          (* --------------- too many rectangles: restore the region *)
          count := count0;
          rect := rect0;
          RETURN FALSE;
        END;
      END;
      (* ------------- normal end *)
      RETURN TRUE;
    END MinusRectUpdate;

  PROCEDURE MinusRectNew(rgn: REGION; r: Geo.Rectangle; VAR rgnRes: REGION): BOOLEAN;
    VAR
      newRgn: REGION;
      j: CARDINAL;
    BEGIN
      IF DrvMem.Alloc(newRgn, SIZE(RegionDescriptor)) THEN
        WITH newRgn^ DO
          (* ---------------------- compute the difference *)
          count := 0;
          IF rgn^.count > 0 THEN
            FOR j := 0 TO rgn^.count - 1 DO
              count := Geo.RectMinusRect(rgn^.rect[j], r, rect, count, MaxRectsNmb - 1);
            END;
          END;
          (* ------------------- test the resulting region - TEMPORARILY !!! *)
          IF count > MaxRectsNmb THEN
            (* ------------- too many rectangles: dispose the new region *)
            DrvMem.Free(newRgn, SIZE(RegionDescriptor));
            RETURN FALSE;
          END;
          (* ------------ include the region into the region chain *)
          next := topRgn;
        END;
        (* ---------------- normal end *)
        topRgn := newRgn;
        rgnRes := newRgn;
        RETURN TRUE;
      ELSE
        (* ------- fail due to heap limitations *)
        RETURN FALSE;
      END;
    END MinusRectNew;

  PROCEDURE ToRects(rgn: REGION; VAR res: ARRAY OF Geo.Rectangle);
    VAR
      j: CARDINAL;
    BEGIN
      WITH rgn^ DO
        IF count > 0 THEN
          FOR j := 0 TO count - 1 DO
            IF j <= HIGH(res) THEN
              res[j] := rect[j];
            END;
          END;
        END;
      END;
    END ToRects;

  PROCEDURE Exist(rgn: REGION): BOOLEAN;
    BEGIN
      RETURN SrvObj.Exist(SrvObj.OBJECT(topRgn), SrvObj.OBJECT(rgn));
    END Exist;

  PROCEDURE Dispose(VAR rgn: REGION);
    VAR
      prevRgn: REGION;
    BEGIN
      prevRgn := REGION(SrvObj.Prev(SrvObj.OBJECT(topRgn), SrvObj.OBJECT(rgn)));
      IF prevRgn # NIL THEN
        (* -------------------- exclude the region from the region chain *)
        IF rgn # topRgn THEN
          prevRgn^.next := rgn^.next;
        ELSE
          topRgn := rgn^.next;
        END;
        (* ------------ deallocate the region descriptor *)
        DrvMem.Free(rgn, SIZE(RegionDescriptor));
      END;
    END Dispose;

  BEGIN
    topRgn := NIL;
  END GeoRgn.
